from fiscalsim_us.model_api import *
from numpy import round


class ar_lump_sum_dist_tax(Variable):
    "Line 31 of form AR1000F"
    value_type = float
    entity = TaxUnit
    label = "Arkansas Lump Sum Distribution Tax"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000TD_LumpSumDistributionAveraging.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        # def high_income_reduction(income):

        #     full_reduction= parameters(period).gov.states.ar.tax.income.high_income_reduction.high_income_reduction_amount
        #     min_income = parameters(period).gov.states.ar.tax.income.rates.regular_bracket_max + 1
        #     phaseout_rate = parameters(period).gov.states.ar.tax.income.high_income_reduction.high_income_reduction_phaseout

        #     def round_to_nearest_50(num):
        #         # Calculate the nearest multiple of 100
        #         nearest_multiple_of_100 = round(num / 100,0) * 100

        #         # Get the last two digits
        #         last_two_digits = num % 100

        #         # Determine the closest ending in "50"
        #         if last_two_digits <= 50 and last_two_digits >= 1:
        #             rounded_income = nearest_multiple_of_100 + 50
        #             return rounded_income
        #         else:
        #             rounded_income = nearest_multiple_of_100 - 50
        #             return rounded_income

        #     std_ded = tax_unit("ar_standard_deduction", period)
        #     itm_ded = tax_unit("ar_itemized_deductions", period)
        #     deduction = where(itm_ded > std_ded, itm_ded, std_ded)

        #     income_less_ded = income - deduction

        #     rounded_income = round_to_nearest_50(income_less_ded)
        #     rounded_min_income = round_to_nearest_50(min_income)

        #     # Calculate the phaseout reduction for each $100 over min_income
        #     excess_income = rounded_income - rounded_min_income
        #     phaseout_reduction = where( excess_income % 100 == 0, (excess_income // 100) * phaseout_rate, ((excess_income // 100) + 1) * phaseout_rate)

        #     # Reduce the credit amount based on the phaseout reduction
        #     reduction_amount = full_reduction - phaseout_reduction

        #     # Ensure credit_amount does not go below 0
        #     reduction_amount = where(reduction_amount < 0 or excess_income < 0 ,
        #         0, reduction_amount)

        #     reduction_amount = round(reduction_amount,0)

        #     return reduction_amount

        p = parameters(period).gov.states.ar.tax.income.lump_sum_dist
        high_income_threshold = (
            parameters(
                period
            ).gov.states.ar.tax.income.rates.regular_bracket_max
            + 1
        )
        high_income_reduction = parameters(
            period
        ).gov.states.ar.tax.income.high_income_reduction.high_income_reduction

        income = tax_unit("ar_distribution_income", period)
        actuarial_value = tax_unit("ar_actuarial_value", period)
        taxable_dist = income + actuarial_value
        min_allowance_threshold = p.min_allowance_threshold

        std_ded = tax_unit("ar_standard_deduction", period)
        itm_ded = tax_unit("ar_itemized_deductions", period)
        deduction = where(itm_ded > std_ded, itm_ded, std_ded)

        line_4_max = p.min_allowance_multiple1_max
        line_4_multiple = p.min_allowance_multiple1

        line_4 = min_(line_4_max, taxable_dist * line_4_multiple)

        min_allowance_subtraction = p.min_allowance_subtract

        line_5 = max_(0, taxable_dist - min_allowance_subtraction)

        line_6_multiple = p.min_allowance_multiple2

        line_6 = line_5 * line_6_multiple

        min_allowance = where(
            taxable_dist < min_allowance_threshold, line_4 - line_6, 0
        )

        line_8 = where(
            taxable_dist < min_allowance_threshold,
            taxable_dist - min_allowance,
            taxable_dist,
        )

        line_9_multiple = p.Line9_multiple

        line_9 = line_8 * line_9_multiple

        # line_9_reduction = high_income_reduction(line_9)
        line_9_reduction = high_income_reduction.calc(line_9)

        line_9_tax = round_(
            where(
                line_9 <= high_income_threshold,
                parameters(period).gov.states.ar.tax.income.rates.rates.calc(
                    (line_9 - deduction)
                ),
                parameters(
                    period
                ).gov.states.ar.tax.income.rates.high_income_rates.calc(
                    (line_9 - deduction)
                ),
            )
            - line_9_reduction,
            0,
        )

        line_11_multiple = p.Line11_multiple

        line_11 = line_9_tax * line_11_multiple

        line_12 = actuarial_value / taxable_dist

        line_13 = min_allowance * line_12

        line_14 = actuarial_value - line_13

        line_15_multiple = p.Line15_multiple

        line_15 = line_14 * line_15_multiple

        line_15_reduction = high_income_reduction.calc(line_15)

        line_15_tax = round_(
            where(
                line_15 <= high_income_threshold,
                parameters(period).gov.states.ar.tax.income.rates.rates.calc(
                    line_15 - deduction
                ),
                parameters(
                    period
                ).gov.states.ar.tax.income.rates.high_income_rates.calc(
                    line_15 - deduction
                ),
            )
            - line_15_reduction,
            0,
        )

        line_17_multiple = p.Line17_multiple

        line_17 = line_15_tax * line_17_multiple

        tax = where(actuarial_value > 0, line_11 - line_17, 0)

        return tax



from fiscalsim_us.model_api import *


class ms_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80108228.pdf#page=1"
        "https://www.law.cornell.edu/regulations/mississippi/35-Miss-Code-R-SS-3-02-11-103"
    )
    defined_for = StateCode.MS

    def formula(tax_unit, period, parameters):
        itemized = tax_unit("ms_itemized_deductions", period)
        standard = tax_unit("ms_standard_deduction", period)
        return max_(itemized, standard)



from fiscalsim_us.model_api import *


class de_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://delcode.delaware.gov/title30/c011/sc02/index.html title 30, chapter 11, subchapter II, section 1108"
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.de.tax.income.deductions.standard
        return p.amount[filing_status]



from fiscalsim_us.model_api import *


class has_marketplace_health_coverage(Variable):
    value_type = bool
    entity = Person
    label = "Receives health insurance from a Marketplace plan."
    definition_period = YEAR
    default_value = False



from fiscalsim_us.data.storage import STORAGE_FOLDER
from policyengine_core.data import Dataset
from fiscalsim_us.data.datasets.cps.cps import CPS_2023


class CalibratedCPS_2023(Dataset):
    name = "calibrated_cps_2023"
    label = "Calibrated CPS (2023)"
    file_path = STORAGE_FOLDER / "calibrated_cps.h5"
    data_format = Dataset.ARRAYS
    time_period = "2023"

    def generate(self):
        from .calibrate import calibrate

        new_data = {}
        cps = CPS_2023()
        cps_data = cps.load()
        adjusted_weights = calibrate(
            "cps_2023",
            time_period="2023",
            training_log_path="calibration_log_cps.csv.gz",
        )
        for variable in cps.variables:
            if variable == "household_weight":
                new_data[variable] = adjusted_weights
            elif "_weight" not in variable:
                new_data[variable] = cps_data[variable][...]

        self.save_dataset(new_data)



from fiscalsim_us.model_api import *


class mn_other_nonrefundable_credits(Variable):
    """
    Other nonrefundable credits that haven't been implemented in other vars.
    Other nonrefundable credits from Lines 2-20 on Minnesota 2023 on Form M1C
    Indivual Tax Return from M1 not yet implemented
    * Credit for long term care insurcance premiums paid
    * Credit for taxes paid to another state
    * Credit for past military service
    * Employer Transit Pass Credit
    * SEED Capital Investment Credit
    * Education Savings Account Contribution Credit
    * Credit for attaining masters degree in teacher licensure field
    * Student Loan Credit
    * Beginning Farmer Management Credit
    * Film Production Credit
    * Tax Credit for Owners of Agricultural Assets
    * Credit for Sales of Manufactured Home Parks to Cooperatives
    * Short Line Railroad Infrastructure Modernization Credit
    * Housing Tax Credit
    * Credit for increasing research activities
    * Carryforward of previous year Farmer Management Credits
    * Carryforward of previous year Owners of Agricultural Assets Credits
    * Carryforward of prior year increasing research activities credit
    * Alternative Minimum Tax Credit

    If another variable gets added, remove from the list.
    """

    value_type = float
    entity = TaxUnit
    label = "Minnesota other nonrefundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1c_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1c_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2024-01/m1c-23.pdf"
    )
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class ctc_value(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC value"
    unit = USD
    documentation = "Actual value of the Child Tax Credit"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return min_(
            tax_unit("ctc", period),
            tax_unit("ctc_limiting_tax_liability", period),
        )



from fiscalsim_us.model_api import *


class mt_itemized_deduction(Variable):
    """
    See itemized deduction schedule on page 7 of the Montana individual tax return form 2 for how to calculate
    Line 15 on Montana individual tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "MT itemized deduction"
    unit = USD
    definition_period = YEAR
    documentation = "Montana itemized deduction"
    reference = (
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/05/Montana-Idividiual-Income-Tax-Return-Form-2-2022v6.2.pdf"
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2022/12/Form-2-2022-Instructions.pdf"
    )
    defined_for = StateCode.MT

    adds = "gov.states.mt.tax.income.deductions.itemized.itemized_deductions"



from fiscalsim_us.model_api import *


class me_non_refundable_child_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "non-refundable ME Child Care Credit"
    unit = USD
    documentation = (
        "The portion of the ME Child Care Credit that is non-refundable."
    )
    reference = "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_a_ff.pdf"
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5218.html"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        me_child_care_credit = tax_unit("me_child_care_credit", period)
        me_refundable_child_care_credit = tax_unit(
            "me_refundable_child_care_credit", period
        )

        return max_(me_child_care_credit - me_refundable_child_care_credit, 0)



from fiscalsim_us.data.storage import STORAGE_FOLDER
from policyengine_core.data import Dataset


class EnhancedCPS_2023(Dataset):
    name = "enhanced_cps_2023"
    label = "Enhanced CPS (2023)"
    file_path = STORAGE_FOLDER / "enhanced_cps.h5"
    data_format = Dataset.ARRAYS
    time_period = "2023"
    url = "release://policyengine/policyengine-us/enhanced-cps-2023/enhanced_cps.h5"

    def generate(self):
        from .puf_extended_cps import PUFExtendedCPS_2023
        from .calibrate import calibrate

        new_data = {}
        cps = PUFExtendedCPS_2023()
        cps_data = cps.load()
        adjusted_weights = calibrate(
            "puf_extended_cps_2023",
            time_period="2023",
        )
        for variable in cps.variables:
            if variable == "household_weight":
                new_data[variable] = adjusted_weights
            elif "_weight" not in variable:
                new_data[variable] = cps_data[variable][...]

        self.save_dataset(new_data)



from fiscalsim_us.model_api import *


class ar_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas adjusted gross income"
    defined_for = StateCode.AR
    unit = USD
    definition_period = YEAR
    # Arkansas Code
    reference = "https://law.justia.com/codes/arkansas/2019/title-26/subtitle-5/chapter-51/subchapter-4/section-26-51-403/"

    adds = ["ar_total_income", "ar_capital_loss_adjustment"]
    subtracts = ["ar_income_adjustments"]



from fiscalsim_us.model_api import *


class va_calc_line_8(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA interim step to get to total virginia agi, line 8 on form 760"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        age_deduction = tax_unit("va_age_deduction", period)

        subtractions_to_federal_agi = tax_unit(
            "va_subtractions_to_federal_agi", period
        )

        return age_deduction + subtractions_to_federal_agi



from io import BytesIO
from zipfile import ZipFile
from policyengine_core.data import Dataset
import pandas as pd
import requests
from tqdm import tqdm
from fiscalsim_us.data.storage import STORAGE_FOLDER

TAX_UNIT_COLUMNS = [
    "ACTC_CRD",
    "AGI",
    "CTC_CRD",
    "EIT_CRED",
    "FEDTAX_AC",
    "FEDTAX_BC",
    "MARG_TAX",
    "STATETAX_A",
    "STATETAX_B",
    "TAX_INC",
]

SPM_UNIT_COLUMNS = [
    "ACTC",
    "CAPHOUSESUB",
    "CAPWKCCXPNS",
    "CHILDCAREXPNS",
    "CHILDSUPPD",
    "EITC",
    "ENGVAL",
    "EQUIVSCALE",
    "FAMTYPE",
    "FEDTAX",
    "FEDTAXBC",
    "FICA",
    "GEOADJ",
    "HAGE",
    "HHISP",
    "HMARITALSTATUS",
    "HRACE",
    "MEDXPNS",
    "NUMADULTS",
    "NUMKIDS",
    "NUMPER",
    "POOR",
    "POVTHRESHOLD",
    "RESOURCES",
    "SCHLUNCH",
    "SNAPSUB",
    "STTAX",
    "TENMORTSTATUS",
    "TOTVAL",
    "WCOHABIT",
    "WEIGHT",
    "WFOSTER22",
    "WICVAL",
    "WKXPNS",
    "WNEWHEAD",
    "WNEWPARENT",
    "WUI_LT15",
    "ID",
]
SPM_UNIT_COLUMNS = ["SPM_" + column for column in SPM_UNIT_COLUMNS]
PERSON_COLUMNS = [
    "PH_SEQ",
    "PF_SEQ",
    "P_SEQ",
    "TAX_ID",
    "SPM_ID",
    "A_FNLWGT",
    "A_LINENO",
    "A_SPOUSE",
    "A_AGE",
    "A_SEX",
    "PEDISEYE",
    "MRK",
    "WSAL_VAL",
    "INT_VAL",
    "SEMP_VAL",
    "FRSE_VAL",
    "DIV_VAL",
    "RNT_VAL",
    "SS_VAL",
    "UC_VAL",
    "ANN_VAL",
    "PNSN_VAL",
    "OI_OFF",
    "OI_VAL",
    "CSP_VAL",
    "PAW_VAL",
    "SSI_VAL",
    "RETCB_VAL",
    "CAP_VAL",
    "WICYN",
    "VET_VAL",
    "WC_VAL",
    "DIS_VAL1",
    "DIS_VAL2",
    "CHSP_VAL",
    "PHIP_VAL",
    "MOOP",
    "PEDISDRS",
    "PEDISEAR",
    "PEDISOUT",
    "PEDISPHY",
    "PEDISREM",
    "PEPAR1",
    "PEPAR2",
    "DIS_SC1",
    "DIS_SC2",
    "PRDTRACE",
    "PRDTHSP",
    "A_MARITL",
]


class RawCPS(Dataset):
    name = "raw_cps"
    label = "Raw CPS"
    time_period = None
    data_format = Dataset.TABLES

    def generate(self) -> pd.DataFrame:
        """Generates the raw CPS dataset."""
        # Files are named for a year after the year the survey represents.
        # For example, the 2020 CPS was administered in March 2021, so it's
        # named 2021.
        file_year = int(self.time_period) + 1
        file_year_code = str(file_year)[-2:]

        CPS_URL_BY_YEAR = {
            2020: "https://www2.census.gov/programs-surveys/cps/datasets/2021/march/asecpub21csv.zip",
            2021: "https://www2.census.gov/programs-surveys/cps/datasets/2022/march/asecpub22csv.zip",
            2022: "https://www2.census.gov/programs-surveys/cps/datasets/2023/march/asecpub23csv.zip",
        }

        if self.time_period not in CPS_URL_BY_YEAR:
            raise ValueError(
                f"No raw CPS data URL known for year {self.time_period}."
            )

        url = CPS_URL_BY_YEAR[self.time_period]

        response = requests.get(url, stream=True)
        total_size_in_bytes = int(
            response.headers.get("content-length", 200e6)
        )
        progress_bar = tqdm(
            total=total_size_in_bytes,
            unit="iB",
            unit_scale=True,
            desc="Downloading ASEC",
        )
        if response.status_code == 404:
            raise FileNotFoundError(
                "Received a 404 response when fetching the data."
            )
        try:
            with (
                BytesIO() as file,
                pd.HDFStore(self.file_path, mode="w") as storage,
            ):
                content_length_actual = 0
                for data in response.iter_content(int(1e6)):
                    progress_bar.update(len(data))
                    content_length_actual += len(data)
                    file.write(data)
                progress_bar.set_description("Downloaded ASEC")
                progress_bar.total = content_length_actual
                progress_bar.close()
                zipfile = ZipFile(file)
                with zipfile.open(f"pppub{file_year_code}.csv") as f:
                    storage["person"] = pd.read_csv(
                        f,
                        usecols=PERSON_COLUMNS
                        + SPM_UNIT_COLUMNS
                        + TAX_UNIT_COLUMNS,
                    ).fillna(0)
                    person = storage["person"]
                with zipfile.open(f"ffpub{file_year_code}.csv") as f:
                    person_family_id = person.PH_SEQ * 10 + person.PF_SEQ
                    family = pd.read_csv(f).fillna(0)
                    family_id = family.FH_SEQ * 10 + family.FFPOS
                    family = family[family_id.isin(person_family_id)]
                    storage["family"] = family
                with zipfile.open(f"hhpub{file_year_code}.csv") as f:
                    person_household_id = person.PH_SEQ
                    household = pd.read_csv(f).fillna(0)
                    household_id = household.H_SEQ
                    household = household[
                        household_id.isin(person_household_id)
                    ]
                    storage["household"] = household
                storage["tax_unit"] = RawCPS._create_tax_unit_table(person)
                storage["spm_unit"] = RawCPS._create_spm_unit_table(person)
        except Exception as e:
            raise ValueError(
                f"Attempted to extract and save the CSV files, but encountered an error: {e} (removed the intermediate dataset)."
            )

    @staticmethod
    def _create_tax_unit_table(person: pd.DataFrame) -> pd.DataFrame:
        tax_unit_df = person[TAX_UNIT_COLUMNS].groupby(person.TAX_ID).sum()
        tax_unit_df["TAX_ID"] = tax_unit_df.index
        return tax_unit_df

    @staticmethod
    def _create_spm_unit_table(person: pd.DataFrame) -> pd.DataFrame:
        return person[SPM_UNIT_COLUMNS].groupby(person.SPM_ID).first()


class RawCPS_2020(RawCPS):
    time_period = 2020
    name = "raw_cps_2020"
    label = "Raw CPS 2020"
    file_path = STORAGE_FOLDER / "raw_cps_2020.h5"


class RawCPS_2021(RawCPS):
    time_period = 2021
    name = "raw_cps_2021"
    label = "Raw CPS 2021"
    file_path = STORAGE_FOLDER / "raw_cps_2021.h5"


class RawCPS_2022(RawCPS):
    time_period = 2022
    name = "raw_cps_2022"
    label = "Raw CPS 2022"
    file_path = STORAGE_FOLDER / "raw_cps_2022.h5"



from fiscalsim_us.model_api import *


class mn_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-12/m1-23.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2024-02/m1-inst-23.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        std_ded = tax_unit("mn_standard_deduction", period)
        itm_ded = tax_unit("mn_itemized_deductions", period)
        itemizing = tax_unit("mn_itemizing", period)
        return where(itemizing, itm_ded, std_ded)



from fiscalsim_us.model_api import *


class wv_low_income_earned_income_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia low-income earned income exclusion"
    defined_for = StateCode.WV
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class ar_other_income(Variable):
    "AR-OI form"
    """
    Additions to income include:
    Federal Depreciation
    HSA/MSA taxable distributions
    Long-term care insurance contracts
    Gambling Winnings
    Lottery/Contest Winnings
    Scholarships/fellowships/stipends
    Pass Through Entity Adjustment

    Subtractions from Income include:
    State Depreciation
    Net operating loss
    Foreign earned income exclusion
    Pass Through Entity Adjustment
    Other
    """

    value_type = float
    entity = Person
    label = "Arkansas other income"
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=10"
    defined_for = StateCode.AR
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class mn_income_tax(Variable):
    """
    Minnesota income tax is either line 24 (if negative tax liability, refund)
    or line 26 (if positive tax liability, payment) on form M1 (2023).
    """

    value_type = float
    entity = TaxUnit
    label = "Minnesota income tax"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-12/m1-23.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2024-02/m1-inst-23.pdf"
    )
    defined_for = StateCode.MN
    adds = ["mn_income_tax_before_refundable_credits"]
    subtracts = ["mn_refundable_credits"]



from fiscalsim_us.model_api import *


class ut_subtractions_from_income(Variable):
    """
    Line 8 on Utah 2022 Individual Income Tax return form TC-40. These
    subtractions from income include the following categories and are listed at
    https://incometax.utah.gov/subtractions with their accompanying codes.
    * 71: Interest from Utah municipal bonds and U. S. Government obligations
    * 77: Native American income
    * 78: Railroad retirement income
    * 79: Equitable adjustments
    * 82: Nonresident active duty military pay
    * 85: State tax refund distributed to beneficiary of trust
    * 88: Nonresident military spouse income
    * 89: FDIC premiums
    * 90: Qualified retirement plan distributions
    * SA: COVID-19 Utah grant funds included in AGI
    """

    value_type = float
    entity = TaxUnit
    label = "UT subtractions from income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



from fiscalsim_us.model_api import *


class self_employment_tax_ald(Variable):
    value_type = float
    entity = TaxUnit
    label = "Self-employment tax ALD deduction"
    unit = USD
    documentation = "Above-the-line deduction for self-employment tax"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/164#f"

    adds = ["self_employment_tax_ald_person"]



from fiscalsim_us.model_api import *


class ut_nonapportionable_nonrefundable_credits(Variable):
    """
    Line 26 on Utah 2022 Individual Income Tax return form TC-40. These
    credits include the following categories and are listed at
    https://incometax.utah.gov/credits with their accompanying codes.
    * 01: At-home parent credit
    * 02: Qualified sheltered workshop cash contribution credit
    * 06: Historic preservation credit
    * 08: Low-income housing credit
    * 12: Credit for increasing research activities in Utah
    * 13: Carryforward of credit for machinery and equipment used to conduct research
    * 17: Credit for income tax paid to another state
    * 19: Live organ donation expenses credit
    * 21: Renewable residential energy systems credit
    * 25: Combat related death credit
    * 27: Veteran employment credit
    * 28: Employing persons who are homeless credit
    * 63: Achieving a Better Life Experience (ABLE) program credit
    * AA: Military survivor benefits credit
    * AG: Special Needs Opportunity Scholarship Program credit
    * AP: Pass-through entity taxpayer income tax credit
    """

    value_type = float
    entity = TaxUnit
    label = "UT nonapportionable nonrefundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



from fiscalsim_us.model_api import *


class va_amount_of_prior_year_overpayment_applied_to_current_year(Variable):
    value_type = float
    entity = TaxUnit
    label = "va tax witheld from the tax year line 21 on form 760 https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2021-760-instructions.pdf for information"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA



from fiscalsim_us.model_api import *


class household_tax_before_refundable_credits(Variable):
    value_type = float
    entity = Household
    label = "tax"
    documentation = "Total tax liability before refundable credits."
    unit = USD
    definition_period = YEAR
    adds = [
        "employee_payroll_tax",
        "self_employment_tax",
        "income_tax_before_refundable_credits",  # Federal.
        "ar_income_tax_before_refundable_credits",
        "ca_income_tax_before_refundable_credits",
        "co_income_tax_before_refundable_credits",
        "dc_income_tax_before_refundable_credits",
        "ia_income_tax_before_refundable_credits",
        "il_total_tax",
        "in_income_tax_before_refundable_credits",
        "ks_income_tax_before_refundable_credits",
        "ky_income_tax_before_refundable_credits",
        "la_income_tax_before_refundable_credits",
        "me_income_tax_before_refundable_credits",
        "ma_income_tax_before_refundable_credits",
        "md_income_tax_before_refundable_credits",
        "mn_income_tax_before_refundable_credits",
        "mt_income_tax_before_refundable_credits",
        "mo_income_tax_before_refundable_credits",
        "nc_income_tax",  # NC has no refundable credits.
        "nd_income_tax_before_refundable_credits",
        "ne_income_tax_before_refundable_credits",
        "nh_income_tax_before_refundable_credits",
        "nj_income_tax_before_refundable_credits",
        "nm_income_tax_before_refundable_credits",
        "ny_income_tax_before_refundable_credits",
        "ok_income_tax_before_refundable_credits",
        "or_income_tax_before_refundable_credits",
        "pa_income_tax",  # PA has no refundable credits.
        "ri_income_tax_before_refundable_credits",
        "sc_income_tax_before_refundable_credits",
        "ut_income_tax_before_refundable_credits",
        "va_income_tax_before_refundable_credits",
        "vt_income_tax_before_refundable_credits",
        "wa_income_tax_before_refundable_credits",
        "wi_income_tax_before_refundable_credits",
        "flat_tax",
        "nyc_income_tax_before_refundable_credits",
    ]

    def formula(household, period, parameters):
        added_components = household_tax_before_refundable_credits.adds
        params = parameters(period)
        flat_tax = params.gov.contrib.ubi_center.flat_tax
        if flat_tax.abolish_payroll_tax:
            added_components = [
                c for c in added_components if c != "employee_payroll_tax"
            ]
        if flat_tax.abolish_federal_income_tax:
            added_components = [
                c
                for c in added_components
                if c != "income_tax_before_refundable_credits"
            ]
        if params.simulation.reported_state_income_tax:
            added_components = [
                "employee_payroll_tax",
                "self_employment_tax",
                "income_tax_before_refundable_credits",  # Federal.
                "flat_tax",
                "spm_unit_state_tax_reported",
            ]
        return add(household, period, added_components)



from fiscalsim_us.model_api import *
import numpy as np


class ar_high_income_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas high income reduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_TaxTables.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        reduction = parameters(
            period
        ).gov.states.ar.tax.income.high_income_reduction.high_income_reduction
        min_income = (
            parameters(
                period
            ).gov.states.ar.tax.income.rates.regular_bracket_max
            + 1
        )
        agi = tax_unit("ar_agi", period)

        def round_to_nearest_50(num):
            # Calculate the nearest multiple of 100
            nearest_multiple_of_100 = np.round(num / 100, 0) * 100

            # Get the last two digits
            last_two_digits = num % 100

            # Determine the closest ending in "50"
            if np.isscalar(num):
                if last_two_digits <= 50 and last_two_digits >= 1:
                    rounded_income = nearest_multiple_of_100 + 50
                    return rounded_income
                else:
                    rounded_income = nearest_multiple_of_100 - 50
                    return rounded_income
            else:
                rounded_income = np.zeros_like(num)
                rounded_income[
                    (last_two_digits <= 50) & (last_two_digits >= 1)
                ] = (
                    nearest_multiple_of_100[
                        (last_two_digits <= 50) & (last_two_digits >= 1)
                    ]
                    + 50
                )
                rounded_income[
                    (last_two_digits > 50) | (last_two_digits < 1)
                ] = (
                    nearest_multiple_of_100[
                        (last_two_digits > 50) | (last_two_digits < 1)
                    ]
                    - 50
                )
                return rounded_income

        std_ded = tax_unit("ar_standard_deduction", period)
        itm_ded = tax_unit("ar_itemized_deductions", period)
        deduction = where(itm_ded > std_ded, itm_ded, std_ded)

        agi_less_ded = agi - deduction

        rounded_income = round_to_nearest_50(agi_less_ded)
        rounded_min_income = round_to_nearest_50(min_income)

        reduction_amount = reduction.calc(rounded_income, right=True)

        return reduction_amount



from fiscalsim_us.model_api import *


class is_wic_at_nutritional_risk(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Assessed as being at nutritional risk for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "At nutritional risk for WIC"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#b_8"

    def formula(person, period, parameters):
        wic_reported = person("receives_wic", period)
        meets_income_test = person.spm_unit("meets_wic_income_test", period)
        meets_categorical_test = person(
            "meets_wic_categorical_eligibility", period
        )
        category = person("wic_category", period)
        risk = parameters(period).gov.usda.wic.nutritional_risk
        imputed_risk = (random(person) < risk[category]) & (
            meets_income_test | meets_categorical_test
        )
        return wic_reported | imputed_risk



from fiscalsim_us.model_api import *


class household_refundable_tax_credits(Variable):
    value_type = float
    entity = Household
    label = "refundable tax credits"
    definition_period = YEAR
    unit = USD
    adds = [
        "income_tax_refundable_credits",  # Federal
        "ar_refundable_credits",  # Arkansas
        "ca_refundable_credits",  # California
        "co_refundable_credits",  # Colorado
        "dc_refundable_credits",  # District of Columbia
        "ia_refundable_credits",  # Iowa
        "il_refundable_credits",  # Illinois
        "in_refundable_credits",  # Indiana
        "ks_refundable_credits",  # Kansas
        "ky_refundable_credits",  # Kentucky
        "la_refundable_credits",  # Louisiana
        "ma_refundable_credits",  # Massachusetts
        "me_refundable_credits",  # Maine
        "md_refundable_credits",  # Maryland
        "mn_refundable_credits",  # Minnesota
        "mt_refundable_credits",  # Montana
        "mo_refundable_credits",  # Missouri
        # Skip NC, which has no refundable credits
        "nd_refundable_credits",  # North Dakota
        "ne_refundable_credits",  # Nebraska
        "nh_refundable_credits",  # New Hampshire
        "nj_refundable_credits",  # New Jersey
        "nm_refundable_credits",  # New Mexico
        "ny_refundable_credits",  # New York
        "ok_refundable_credits",  # Oklahoma
        "or_refundable_credits",  # Oregon
        # Skip PA, which has no refundable credits
        "ri_refundable_credits",  # Rhode Island
        "sc_refundable_credits",  # South Carolina
        "ut_refundable_credits",  # Utah
        "va_refundable_credits",  # Virginia
        "vt_refundable_credits",  # Vermont
        "wa_refundable_credits",  # Washington
        "wi_refundable_credits",  # Wisconsin
        # LOCAL
        "nyc_refundable_credits",  # New York City
    ]

    def formula(household, period, parameters):
        added_components = household_refundable_tax_credits.adds
        params = parameters(period)
        p = params.gov.contrib.ubi_center.flat_tax
        if p.abolish_federal_income_tax:
            added_components = [
                c
                for c in added_components
                if c != "income_tax_refundable_credits"
            ]
        if params.simulation.reported_state_income_tax:
            added_components = [
                "income_tax_refundable_credits",  # Federal.
            ]
        return add(household, period, added_components)



from fiscalsim_us.model_api import *


class ar_stillborn_child_credit(Variable):
    "Line 5 of form AR1000TC"
    value_type = float
    entity = TaxUnit
    label = "Arkansas Stillborn Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000TC_Schedule_of_TaxCredits_andBusiness_IncentiveCredits.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ar.tax.income.credits.other_credits

        amount = p.stillbirth_credit

        num = tax_unit("tax_unit_stillborn_children", period)

        return amount * num



from fiscalsim_us.model_api import *


class mn_income_tax_before_credits(Variable):
    """
    Line 15 from form M1 (2023)
    """

    value_type = float
    entity = TaxUnit
    label = "Minnesota income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2024-02/m1-inst-23.pdf"
    )
    defined_for = StateCode.MN
    adds = ["mn_basic_tax", "mn_amt"]



from fiscalsim_us.model_api import *


class ar_total_income(Variable):
    "Line 23 of AR1000F"
    value_type = float
    entity = Person
    label = "Total income"
    unit = USD
    documentation = "Total income, as defined in the Arkansas state Tax Form."
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        sources = parameters(period).gov.states.ar.tax.income.income_sources
        ira_exemption = parameters(
            period
        ).gov.states.ar.tax.income.rates.ira_exemption
        retirement_sources = parameters(
            period
        ).gov.states.ar.tax.income.retirement_sources
        total = 0
        not_dependent = ~person("is_tax_unit_dependent", period)
        for source in sources:
            amount = person(source, period)
            total += where(
                source in retirement_sources,
                not_dependent * max_(0, amount - ira_exemption),
                not_dependent * max_(0, amount),
            )
        return total



from fiscalsim_us.model_api import *


class ny_agi_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY AGI subtractions"
    unit = USD
    documentation = "Subtractions from NY AGI over federal AGI."
    definition_period = YEAR
    dict(
        title="N.Y. Comp. Codes R. & Regs. tit. 20 § 112.3",
        href="https://casetext.com/regulation/new-york-codes-rules-and-regulations/title-20-department-of-taxation-and-finance/chapter-ii-income-taxes-and-estate-taxes/subchapter-a-new-york-state-personal-income-tax-under-article-22-of-the-tax-law/article-2-residents/part-112-new-york-adjusted-gross-income-of-a-resident-individual/section-1123-modifications-reducing-federal-adjusted-gross-income",
    )

    def formula(tax_unit, period, parameters):
        taxable_ss = add(tax_unit, period, ["taxable_social_security"])
        us_govt_interest = tax_unit("us_govt_interest", period)
        investment_in_529_plan = tax_unit("investment_in_529_plan", period)
        person = tax_unit.members
        pension_income = person("pension_income", period)
        age = person("age", period)

        pension_exclusion = parameters(
            period
        ).gov.states.ny.tax.income.agi.subtractions.pension_exclusion
        meets_age_test = age >= pension_exclusion.min_age
        deductible_pensions = meets_age_test * min_(
            pension_income, pension_exclusion.cap
        )

        return (
            taxable_ss
            + us_govt_interest
            + investment_in_529_plan
            + tax_unit.sum(deductible_pensions)
        )



from fiscalsim_us.model_api import *


class medical_out_of_pocket_expenses(Variable):
    value_type = float
    entity = Person
    label = "Medical out of pocket expenses"
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class is_eitc_qualifying_child(Variable):
    value_type = bool
    entity = Person
    label = "EITC qualifying child"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/152#c"

    def formula(person, period, parameters):
        dependent = person("is_tax_unit_dependent", period)
        disabled = person("is_permanently_and_totally_disabled", period)
        age = person("age", period)
        student = person("is_full_time_student", period)
        max_ages = parameters(
            period
        ).gov.irs.credits.eitc.qualifying_child.max_age
        max_age = where(student, max_ages.student, max_ages.non_student)
        age_qualifies = age <= max_age
        return dependent & (age_qualifies | disabled)



from fiscalsim_us.model_api import *


class mt_agi(Variable):
    """
    Montana adjusted gross income. Line 14 on Montana individual income tax
    return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "MT adjusted gross income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = ["adjusted_gross_income", "mt_income_additions"]
    subtracts = ["mt_income_subtractions"]



from fiscalsim_us.model_api import *


class ssi_amount_if_eligible(Variable):
    value_type = float
    entity = Person
    label = "SSI amount if eligible"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382#b"

    def formula(person, period, parameters):
        ssi = parameters(period).gov.ssa.ssi.amount
        return (
            where(
                person("ssi_claim_is_joint", period),
                ssi.couple,
                ssi.individual,
            )
            * MONTHS_IN_YEAR
        )



from fiscalsim_us.model_api import *


class mn_taxable_income(Variable):
    """
    Line 9 from form M1 (2023)
    """

    value_type = float
    entity = TaxUnit
    label = "Minnesota taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-12/m1-23.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2024-02/m1-inst-23.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        ADDS = ["adjusted_gross_income", "mn_additions"]
        SUBTRACTS = [
            "mn_subtractions",
            "mn_prev_year_state_refund",
            "mn_exemptions",
            "mn_deductions",
        ]

        income = add(tax_unit, period, ADDS) - add(tax_unit, period, SUBTRACTS)
        return max_(0, income)



from fiscalsim_us.model_api import *


class dc_tanf(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF"
    unit = USD
    definition_period = YEAR
    defined_for = "dc_tanf_eligible"

    def formula(spm_unit, period, parameters):
        grant_standard = spm_unit("dc_tanf_grant_standard", period)
        countable_income = spm_unit("dc_tanf_countable_income", period)
        return max_(grant_standard - countable_income, 0)



from fiscalsim_us.model_api import *


class ar_capital_gains(Variable):
    value_type = float
    entity = Person
    label = "Arkansas taxable capital gains"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000D_CapitalGains.pdf"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        # filing_status = person('filing_status', period)

        p = parameters(period).gov.states.ar.tax.income.capital_gains

        max_taxable_gain = p.max_taxable_capital_gain

        ar_lt_taxable_pct = p.long_term_cap_gain_taxable_pct

        lt_gains = person("long_term_capital_gains", period)

        non_d_gains = person("non_sch_d_capital_gains", period)

        lt_dep_adjustment = person("ar_lt_dep_adjustment", period)

        ar_lt_gain = lt_gains + non_d_gains + lt_dep_adjustment

        st_gain = person("short_term_capital_gains", period)

        st_dep_adjustment = person("ar_st_dep_adjustment", period)

        ar_st_gain = st_gain + st_dep_adjustment

        # if there are short term losses, but net gains are positive, net gain should equal the min of max_taxable_gain or the difference between the long term gain and short term loss.
        # if there short term losses, and net losses, net gain should equal the total loss
        # if there are short term gains, those are handled later.

        ar_net_lt_gain = where(
            ar_st_gain < 0,
            where(
                ar_lt_gain + ar_st_gain >= 0,
                min_(max_taxable_gain, ar_lt_gain + ar_st_gain),
                ar_lt_gain + ar_st_gain,
            ),
            min_(ar_lt_gain, max_taxable_gain),
        )

        ar_taxable_amount = where(
            ar_net_lt_gain > 0,
            ar_net_lt_gain * ar_lt_taxable_pct,
            ar_net_lt_gain,
        )

        total_taxable = where(
            ar_st_gain >= 0, ar_st_gain + ar_taxable_amount, ar_taxable_amount
        )

        return total_taxable



"""This file contains the fiscalsim-us package's metadata and dependencies."""

from setuptools import find_packages, setup

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

setup(
    name="fiscalsim-us",
    version="0.3.6",
    author="Richard W. Evans",
    author_email="rick@abundance.institute",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    description="FiscalSim federal and state individual tax and benefit system for the US",
    keywords="tax benefit microsimulation fiscal state household individual personal",
    license="http://www.fsf.org/licensing/licenses/agpl-3.0.html",
    url="https://github.com/TheCGO/fiscalsim-us",
    include_package_data=True,  # Will read MANIFEST.in
    data_files=[
        (
            "share/openfisca/openfisca-country-template",
            ["CHANGELOG.md", "LICENSE", "README.md"],
        ),
    ],
    install_requires=[
        "click==8.1.3",
        "h5py",
        "microdf_python",
        "numpy>=1.24, <1.25",
        "pandas",
        "pathlib",
        "policyengine-core>=2.8,<2.9",
        "pytest",
        "pytest-dependency",
        "pytest-cov",
        "pyyaml",
        "requests",
        "synthimpute",
        "tables",
        "tabulate",
        "tqdm",
    ],
    extras_require={
        "dev": [
            "autopep8",
            "black",
            "coverage",
            "jupyter-book>=0.11.3",
            "jupyter",
            "linecheck",
            "markupsafe",
            "plotly",
            "bokeh>=3.1.1",
            "pydata-sphinx-theme",
            "setuptools",
            "sphinx>=3.5.4",
            "sphinx-argparse",
            "sphinxcontrib-bibtex>=2.0.0",
            "sphinx-math-dollar",
            "wheel",
            "yaml-changelog",
            "survey-enhance",
        ],
    },
    python_requires=">=3.10, <3.12",
    entry_points={
        "console_scripts": [
            "fiscalsim-us = fiscalsim_us.tools.cli:main",
        ],
    },
    packages=find_packages(),
)



from fiscalsim_us.model_api import *


class oh_bonus_depreciation_add_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "OH bonus depreciation add back"
    definition_period = YEAR
    documentation = ""
    reference = ""



from fiscalsim_us.model_api import *


class property_tax_primary_residence(Variable):
    value_type = float
    entity = TaxUnit
    label = "Property tax for the real estate that includes the principal residence"
    unit = USD
    definition_period = YEAR

    adds = ["real_estate_taxes"]



from fiscalsim_us.model_api import *


class ar_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=14"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ar.tax.income.deductions
        filing_status = tax_unit("filing_status", period)
        return p.standard[filing_status]



from fiscalsim_us.model_api import *


class cdcc_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "CDCC credit rate"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21#a_2"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.credits.cdcc
        agi = tax_unit("adjusted_gross_income", period)

        # First phase-out
        excess_agi = max_(0, agi - p.phase_out.start)
        increments = np.ceil(excess_agi / p.phase_out.increment)
        percentage_reduction = increments * p.phase_out.rate
        phased_out_rate = max_(
            p.phase_out.min, p.phase_out.max - percentage_reduction
        )

        # Second phase-out
        second_excess_agi = max_(0, agi - p.phase_out.second_start)
        second_increments = np.ceil(second_excess_agi / p.phase_out.increment)
        second_percentage_reduction = second_increments * p.phase_out.rate
        return max_(0, phased_out_rate - second_percentage_reduction)



from fiscalsim_us.model_api import *


class iitax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "federal income tax"
    documentation = "Total federal individual income tax liability."
    adds = [
        "income_tax_before_credits",
    ]
    subtracts = [
        "income_tax_refundable_credits",
        "income_tax_capped_non_refundable_credits",
    ]

    adds = ["income_tax_before_refundable_credits"]
    subtracts = ["income_tax_refundable_credits"]

    def formula(person, period, parameters):
        if parameters(
            period
        ).gov.contrib.ubi_center.flat_tax.abolish_federal_income_tax:
            return 0
        else:
            added_components = add(person, period, iitax.adds)
            subtracted_components = add(person, period, iitax.subtracts)
            return added_components - subtracted_components


income_tax = variable_alias("income_tax", iitax)



from fiscalsim_us.model_api import *


class housing_cost(Variable):
    value_type = float
    entity = SPMUnit
    label = "Housing cost"
    unit = USD
    definition_period = YEAR

    adds = ["rent"]



from fiscalsim_us.model_api import *


class ar_other_misc_deductions(Variable):
    """
    Line 29 of 2022 AR3, Arkansas Itemized Deductions

    Deductions include:
    Volunteer firefighter expenses
    Gambling Losses
    Other miscellaneous deductions
    """

    value_type = float
    entity = TaxUnit
    label = "AR other miscellaneous deductions"
    unit = USD
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR3_ItemizedDeduction.pdf"
    definition_period = YEAR
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class ny_college_tuition_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY college tuition credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (t)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        tuition = add(tax_unit, period, ["qualified_tuition_expenses"])
        p = parameters(period).gov.states.ny.tax.income.credits.college_tuition
        amount = p.rate.calc(tuition) * p.applicable_percentage
        itemizes = tax_unit("ny_itemizes", period)
        return ~itemizes * amount



from fiscalsim_us.model_api import *


class mt_tax_due(Variable):
    """
    Line 26 on Montana state individual tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "Montana tax due"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT


def formula(tax_unit, period, parameters):
    line20 = tax_unit("mt_tax_after_nonrefundable_credits")
    line25 = tax_unit("mt_total_payments")

    return where(line25 < line20, line20 - line25, 0)



from fiscalsim_us.model_api import *


class va_calc_line_3(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA interim step to get to total virginia agi https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2021-760-instructions.pdf for information"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)

        additions_to_federal_agi = tax_unit(
            "va_additions_to_federal_agi", period
        )

        return federal_agi + additions_to_federal_agi



from fiscalsim_us.model_api import *


class mn_amt_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota alternative minimum tax (AMT) taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1mt_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1mt_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.amt
        fed_agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)

        ADDITIONS = [
            "mn_adjustmnets_fed6251",
            "non_mn_bond_interest",
            "mn_depletion",
            "mn_intangible_drilling_costs",
        ]
        # currently not included:
        # line 4 - other addition from M1MB and M1NC

        line8 = fed_agi + add(tax_unit, period, ADDITIONS)

        SUBTRACTIONS = [
            "mn_medical_dental_deduction",
            "interest_deduction",
            "charitable_deduction",
            "mn_casualty_theft_deduction",
            "mn_unreimbursed_employee_deduction",
            "mn_disabled_impairment_work_deduction",
            "us_govt_interest",
        ]
        # currently not included:
        # line 15 - State income tax refund
        # line 16 - line federal bonus deprecitaion subtraction
        # line 18 - other subtractions

        # line 20
        amt_income_before_std = line8 - add(tax_unit, period, SUBTRACTIONS)

        # line 23
        income_over_phase_out = max_(
            0, amt_income_before_std - p.std_deduct_phase_out[filing_status]
        )

        # line 25
        phased_std_deduct = max_(
            0,
            p.standard_deduct[filing_status]
            - (income_over_phase_out * p.mult),
        )

        return amt_income_before_std - phased_std_deduct



from fiscalsim_us.model_api import *


class ar_capital_loss_adjustment(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas capital loss adjustment"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000D_CapitalGains.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)

        person = tax_unit.members
        not_dependent = ~person("is_tax_unit_dependent", period)

        capital_gains = 0
        capital_gains += not_dependent * add(
            person, period, ["ar_capital_gains"]
        )

        capital_gains = tax_unit.sum(capital_gains)

        cap_loss_cap = parameters(
            period
        ).gov.states.ar.tax.income.capital_gains.capital_loss_cap[
            filing_status
        ]

        adjustment = where(
            capital_gains < -cap_loss_cap, -capital_gains - cap_loss_cap, 0
        )

        return adjustment



from fiscalsim_us.model_api import *


class ma_dependent_or_dependent_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA dependent or dependent care credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.mass.gov/info-details/mass-general-laws-c62-ss-6"  # (x-y)
    )
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        # MA taxpayers can only take either the dependent credit or the
        # dependent care credit.
        return max_(
            tax_unit("ma_dependent_credit", period),
            tax_unit("ma_dependent_care_credit", period),
        )



from fiscalsim_us.model_api import *


class oh_retirement_income_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Ohio Retirement Income Credit Eligible"
    unit = USD
    definition_period = YEAR
    reference = "https://codes.ohio.gov/ohio-revised-code/section-5747.055"
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.credits.retirement_income

        agi = tax_unit("oh_agi", period)
        return agi < p.income_threshold



from fiscalsim_us.model_api import *


class dwks9(Variable):
    value_type = float
    entity = TaxUnit
    label = "DWKS9"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p23250 = add(
            tax_unit,
            period,
            ["long_term_capital_gains", "qualified_dividend_income"],
        )
        c23650 = tax_unit("c23650", period)
        dwks7 = min_(p23250, c23650)  # SchD lines 15 and 16, respectively
        # dwks8 = min(dwks3, dwks4)
        # dwks9 = max(0., dwks7 - dwks8)
        # BELOW TWO STATEMENTS ARE UNCLEAR IN LIGHT OF dwks9=... COMMENT
        e01100 = add(tax_unit, period, ["non_sch_d_capital_gains"])
        c24510 = where(e01100 > 0, e01100, max_(0, dwks7) + e01100)
        return max_(
            0,
            c24510 - min_(0, tax_unit("investment_income_form_4952", period)),
        )



from fiscalsim_us.model_api import *
from numpy import round


class mn_amt(Variable):
    """
    Line 11 of Form M1 (2023) and line 29 from form M1MT (2023). This amount is
    reported as the liability above the basic tax liability calculation.
    """

    value_type = float
    entity = TaxUnit
    label = "MN alternative minimum tax (amount above basic tax)"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1mt_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1mt_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-12/m1mt-23.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.amt
        amt_income = tax_unit("mn_amt_taxable_income", period)
        amt = max_(0, amt_income * p.rate)
        basic_tax = tax_unit("mn_basic_tax", period)

        return round(max_(0, amt - basic_tax))



from fiscalsim_us.model_api import *


class self_employment_social_security_tax(Variable):
    value_type = float
    entity = Person
    label = "Self-employment Social Security tax"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        rate = parameters(period).gov.irs.self_employment.social_security_rate
        income = person(
            "social_security_taxable_self_employment_income", period
        )
        return rate * income



from fiscalsim_us.model_api import *


class snap_standard_deduction(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "Standard deduction for calculating SNAP benefit amount"
    label = "SNAP standard deduction"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_1"

    def formula(spm_unit, period, parameters):
        standard_deductions = parameters(
            period
        ).gov.usda.snap.income.deductions.standard
        state_group = spm_unit.household("state_group_str", period)
        capped_household_size = min_(spm_unit("spm_unit_size", period), 6)
        return standard_deductions[state_group][capped_household_size]



from fiscalsim_us.model_api import *


class hi_military_pay_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii military reserve or national guard duty pay exclusion"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=13"
    )
    definition_period = YEAR
    defined_for = StateCode.HI



from fiscalsim_us.model_api import *


class HUDIncomeLevel(Enum):
    ABOVE_MODERATE = "Above moderate"
    MODERATE = "Moderate"
    LOW = "Low"
    VERY_LOW = "Very low"
    ESPECIALLY_LOW = "Especially low"


class hud_income_level(Variable):
    value_type = Enum
    entity = SPMUnit
    possible_values = HUDIncomeLevel
    default_value = HUDIncomeLevel.ABOVE_MODERATE
    label = "HUD income level"
    documentation = "Income level for HUD programs"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        # Get annual income.
        annual_income = spm_unit("hud_annual_income", period)
        # Get household size.
        size = spm_unit("spm_unit_size", period)
        # Get area median income.
        ami = spm_unit("ami", period)
        # avoid array divide-by-zero warning by not using where() function
        # see following GitHub issue for more details:
        # https://github.com/PolicyEngine/policyengine-us/issues/2496
        income_ami_ratio = np.zeros_like(ami)
        mask = ami != 0
        income_ami_ratio[mask] = annual_income[mask] / ami[mask]
        # Look up thresholds for each income level.
        p = parameters(period).gov.hud.ami_limit
        size_limit = p.family_size
        size_limit_excess = p.per_person_exceeding_4
        size_exceeding_4 = max_(size - 4, 0)
        size_capped_at_4 = min_(size, 4)
        moderate_threshold = (
            size_limit.MODERATE[size_capped_at_4]
            + size_limit_excess.MODERATE * size_exceeding_4
        )
        low_threshold = (
            size_limit.LOW[size_capped_at_4]
            + size_limit_excess.LOW * size_exceeding_4
        )
        very_low_threshold = (
            size_limit.VERY_LOW[size_capped_at_4]
            + size_limit_excess.VERY_LOW * size_exceeding_4
        )
        especially_low_threshold = (
            size_limit.ESPECIALLY_LOW[size_capped_at_4]
            + size_limit_excess.ESPECIALLY_LOW * size_exceeding_4
        )
        # Return the lowest matching one.
        return select(
            [
                income_ami_ratio <= especially_low_threshold,
                income_ami_ratio <= very_low_threshold,
                income_ami_ratio <= low_threshold,
                income_ami_ratio <= moderate_threshold,
                True,
            ],
            [
                HUDIncomeLevel.ESPECIALLY_LOW,
                HUDIncomeLevel.VERY_LOW,
                HUDIncomeLevel.LOW,
                HUDIncomeLevel.MODERATE,
                HUDIncomeLevel.ABOVE_MODERATE,
            ],
        )



from fiscalsim_us.model_api import *


class taxsim_mstat(Variable):
    value_type = int
    entity = TaxUnit
    label = "Marital Status"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        marital_status = tax_unit("mars", period)
        marital = marital_status.possible_values
        return select(
            [
                marital_status == marital.SINGLE,
                marital_status == marital.HEAD_OF_HOUSEHOLD,
                marital_status == marital.JOINT,
                marital_status == marital.SEPARATE,
                marital_status == marital.SURVIVING_SPOUSE,
            ],
            [
                1,
                1,
                2,
                6,
                8,
            ],
        )



from fiscalsim_us.model_api import *
from fiscalsim_us.variables.household.demographic.geographic.county.county_enum import (
    County,
)


class county(Variable):
    value_type = Enum
    possible_values = County
    default_value = County.UNKNOWN
    entity = Household
    label = "County"
    definition_period = YEAR

    def formula(household, period, parameters):
        # Attempt to look up from ZIP code
        zip_code = household("zip_code", period).astype(int)
        zip_codes = ZIP_CODE_DATASET.set_index("zip_code")
        county_name = zip_codes.county[zip_code]
        state_code = zip_codes.state[zip_code]
        county_key = county_name.apply(
            lambda name: name.replace(" ", "_")
            .replace("-", "_")
            .replace(".", "")
            .replace("'", "_")
            .strip()
            .upper()
        )
        county_state = county_key.str.cat(state_code, sep="_")
        county_names = pd.Series(
            np.arange(len(County._member_names_)), index=County._member_names_
        )
        return county_names[county_state]



from fiscalsim_us.model_api import *


class mt_state_local_tax_deduction(Variable):
    """
    Line 5 on itemized deductions schedule
    Includes the following
    *General state and local sales taxes
    *Local income taxes
    *Real estate taxes paid
    *Value-based personal property taxes
    Add each of those together but not more than $10,000 if you are single,
        head of household, or married filing jointly; or $5,000 if you
        are married filing separately
    """

    value_type = float
    entity = TaxUnit
    label = "Montana state and local taxes in prior year"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        salt_amount = add(
            tax_unit,
            period,
            [
                "statelocal_sales_or_prior_inctax",
                "real_estate_taxes",
            ],
        )
        salt = parameters(
            period
        ).gov.state.mt.tax.income.deductions.itemized.misc
        cap = salt.salt_cap[tax_unit("filing_status", period)]
        return min_(cap, salt_amount)



from fiscalsim_us.model_api import *


class mt_subtractions_to_income(Variable):
    """
    Line 13 on Montana 2022 Individual Income Tax return form 2. These
    subtractions to income include the categories that are listed
    in Montana Subtractions Schedule on page 5 of Montana 2022 Individual Income Tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "MT subtractions to income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class amt_income(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "AMT taxable income"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/55#b_2"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("taxable_income", period)
        # Add back excluded deductions
        itemizing = tax_unit("tax_unit_itemizes", period)
        standard_deduction = tax_unit("standard_deduction", period)
        salt_deduction = tax_unit("salt_deduction", period)
        excluded_deductions = where(
            itemizing,
            salt_deduction,
            standard_deduction,
        )
        amt_inc = taxable_income + excluded_deductions
        amt = parameters(period).gov.irs.income.amt
        filing_status = tax_unit("filing_status", period)
        separate_addition = max_(
            0,
            min_(
                amt.exemption.amount[filing_status],
                amt.exemption.phase_out.rate
                * max_(0, amt_inc - amt.exemption.separate_limit),
            ),
        ) * (filing_status == filing_status.possible_values.SEPARATE)
        return amt_inc + separate_addition


class regular_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Regular tax before credits"
    documentation = "Regular tax on regular taxable income before credits"
    unit = USD

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        dwks1 = tax_unit("taxable_income", period)

        capital_gains = parameters(period).gov.irs.capital_gains.brackets

        dwks16 = min_(capital_gains.thresholds["1"][filing_status], dwks1)
        dwks17 = min_(tax_unit("dwks14", period), dwks16)
        dwks20 = dwks16 - dwks17
        lowest_rate_tax = capital_gains.rates["1"] * dwks20
        # Break in worksheet lines
        dwks13 = tax_unit("dwks13", period)
        dwks21 = min_(dwks1, dwks13)
        dwks22 = dwks20
        dwks23 = max_(0, dwks21 - dwks22)
        dwks25 = min_(capital_gains.thresholds["2"][filing_status], dwks1)
        dwks19 = tax_unit("dwks19", period)
        dwks26 = min_(dwks19, dwks20)
        dwks27 = max_(0, dwks25 - dwks26)
        dwks28 = min_(dwks23, dwks27)
        dwks29 = capital_gains.rates["2"] * dwks28
        dwks30 = dwks22 + dwks28
        dwks31 = dwks21 - dwks30
        dwks32 = capital_gains.rates["3"] * dwks31
        # Break in worksheet lines
        dwks33 = min_(
            tax_unit("dwks9", period),
            add(tax_unit, period, ["unrecaptured_section_1250_gain"]),
        )
        dwks10 = tax_unit("dwks10", period)
        dwks34 = dwks10 + dwks19
        dwks36 = max_(0, dwks34 - dwks1)
        dwks37 = max_(0, dwks33 - dwks36)
        dwks38 = 0.25 * dwks37
        # Break in worksheet lines
        dwks39 = dwks19 + dwks20 + dwks28 + dwks31 + dwks37
        dwks40 = dwks1 - dwks39
        dwks41 = 0.28 * dwks40

        # Compute regular tax using bracket rates and thresholds
        reg_taxinc = max_(0, dwks19)
        p = parameters(period).gov.irs.income
        bracket_tops = p.bracket.thresholds
        bracket_rates = p.bracket.rates
        reg_tax = 0
        bracket_bottom = 0
        for i in range(1, len(list(bracket_rates.__iter__())) + 1):
            b = str(i)
            bracket_top = bracket_tops[b][filing_status]
            reg_tax += bracket_rates[b] * amount_between(
                reg_taxinc, bracket_bottom, bracket_top
            )
            bracket_bottom = bracket_top

        # Return to worksheet lines
        dwks42 = reg_tax
        dwks43 = sum(
            [
                dwks29,
                dwks32,
                dwks38,
                dwks41,
                dwks42,
                lowest_rate_tax,
            ]
        )
        dwks44 = tax_unit("income_tax_main_rates", period)
        dwks45 = min_(dwks43, dwks44)
        hasqdivltcg = tax_unit("hasqdivltcg", period)
        return where(hasqdivltcg, dwks45, dwks44)


class alternative_minimum_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) liability"

    def formula(tax_unit, period, parameters):
        amt_income = tax_unit("amt_income", period)
        # Form 6251, Part II top
        amt = parameters(period).gov.irs.income.amt
        phase_out = amt.exemption.phase_out
        filing_status = tax_unit("filing_status", period)
        line29 = max_(
            0,
            (
                amt.exemption.amount[filing_status]
                - phase_out.rate
                * max_(0, amt_income - phase_out.start[filing_status])
            ),
        )
        age_head = tax_unit("age_head", period)
        child = amt.exemption.child
        young_head = (age_head != 0) & (age_head < child.max_age)
        no_or_young_spouse = tax_unit("age_spouse", period) < child.max_age
        line29 = where(
            young_head & no_or_young_spouse,
            min_(line29, tax_unit("filer_earned", period) + child.amount),
            line29,
        )
        line30 = max_(0, amt_income - line29)
        brackets = amt.brackets
        amount_over_threshold = line30 - brackets.thresholds["1"] / tax_unit(
            "sep", period
        )
        line3163 = brackets.rates["1"] * line30 + brackets.rates["2"] * max_(
            0, amount_over_threshold
        )
        dwks10, dwks13, dwks14, dwks19, e24515 = [
            add(tax_unit, period, [variable])
            for variable in [
                "dwks10",
                "dwks13",
                "dwks14",
                "dwks19",
                "unrecaptured_section_1250_gain",
            ]
        ]
        form_6251_part_iii_required = np.any(
            [
                variable > 0
                for variable in [
                    dwks10,
                    dwks13,
                    dwks14,
                    dwks19,
                    e24515,
                ]
            ]
        )

        # Complete Form 6251, Part III

        line37 = dwks13
        line38 = e24515
        line39 = min_(line37 + line38, dwks10)
        line40 = min_(line30, line39)
        line41 = max_(0, line30 - line40)
        amount_over_threshold = max_(
            0, line41 - amt.brackets.thresholds["1"] / tax_unit("sep", period)
        )
        line42 = (
            amt.brackets.rates["1"] * line41
            + amt.brackets.rates["2"] * amount_over_threshold
        )
        line44 = dwks14
        cg = amt.capital_gains.brackets
        line45 = max_(0, cg.thresholds["1"][filing_status] - line44)
        line46 = min_(line30, line37)
        line47 = min_(line45, line46)
        cgtax1 = line47 * cg.rates["1"]
        line48 = line46 - line47
        line51 = dwks19
        line52 = line45 + line51
        line53 = max_(0, cg.thresholds["2"][filing_status] - line52)
        line54 = min_(line48, line53)
        cgtax2 = line54 * cg.rates["2"]
        line56 = line47 + line54
        line57 = where(line41 == line56, 0, line46 - line56)
        linex2 = where(line41 == line56, 0, max_(0, line54 - line48))
        cgtax3 = line57 * cg.rates["3"]
        line61 = where(
            line38 == 0,
            0,
            0.25 * max_(0, (line30 - line41 - line56 - line57 - linex2)),
        )
        line62 = line42 + cgtax1 + cgtax2 + cgtax3 + line61
        line64 = min_(line3163, line62)
        line31 = where(form_6251_part_iii_required, line64, line3163)

        # Form 6251, Part II bottom
        line32 = tax_unit("foreign_tax_credit", period)
        line33 = line31 - line32
        return max_(
            0,
            line33
            - max_(
                0,
                (
                    tax_unit("regular_tax_before_credits", period)
                    - line32
                    - tax_unit("form_4972_lumpsum_distributions", period)
                ),
            ),
        )



from fiscalsim_us.model_api import *


class va_estimated_tax_payments_for_tax_year(Variable):
    value_type = float
    entity = TaxUnit
    label = "va estimated tax payments for the current taxable year line 20 on form 760"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA



from fiscalsim_us.model_api import *


class ar_lt_dep_adjustment(Variable):
    value_type = float
    entity = Person
    label = "Arkansas difference in long term capital gains depreciation from federal amount"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000D_CapitalGains.pdf"
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class md_ctc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Maryland Child Tax Credit"
    definition_period = YEAR
    reference = (
        "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-751-effective-until-712026-tax-credit-for-qualified-child"
        "https://law.justia.com/codes/maryland/2022/tax-general/title-10/subtitle-7/section-10-751/"
    )
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.credits.ctc
        return tax_unit("adjusted_gross_income", period) <= p.agi_cap



from fiscalsim_us.model_api import *


class e07240(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Retirement savings contributions credit from Form 8880"
    unit = USD


retirement_savings_credit = variable_alias("retirement_savings_credit", e07240)



import streamlit as st

st.set_page_config(layout="wide")

st.title("PUF imputation")

import pandas as pd
import numpy as np

puf = pd.read_csv("~/Downloads/puf_2015.csv")
demographics = pd.read_csv("~/Downloads/demographics_2015.csv")

# Variables to match on without demographics:
# - Filing status
# - Adult dependents

from fiscalsim_us import Microsimulation

simulation = Microsimulation()

cps_data = simulation.calculate_dataframe(
    [
        # Variables to match on.
        "filing_status",
        "tax_unit_count_dependents",
        "tax_unit_children",
        "people",
    ]
)

# Apply renames as computations.
# Exemptions: The IRS caps at 3 per type in the PUF.
EXEMPTION_CAP = 3
# XOCAH: Exemptions for Children Living at Home
cps_data["XOCAH"] = np.minimum(cps_data.tax_unit_children, EXEMPTION_CAP)

# XOODEP: Exemptions for Other Dependents.
cps_data["XOODEP"] = np.minimum(
    cps_data.tax_unit_count_dependents - cps_data.tax_unit_children,
    EXEMPTION_CAP,
)

RENAMES = dict(
    filing_status={
        "SINGLE": 1,
        "JOINT": 2,
        "SURVIVING_SPOUSE": 2,
        "SEPARATE": 3,
        "HEAD_OF_HOUSEHOLD": 4,
    }
)

cps_data["MARS"] = cps_data.filing_status.replace(RENAMES["filing_status"])

cps_data["XTOT"] = cps_data.people

cps_data = cps_data[["MARS", "XOCAH", "XOODEP"]]

INPUT_VARIABLES = ["MARS", "XOCAH", "XOODEP"]

OUTPUT_VARIABLES = [
    "E00200",  # Salaries and wages
    "E00300",  # Interest received
    "E00400",  # Tax-exempt interest income
    "E00600",  # Dividends included in AGI
    "E00650",  # Qualified Dividends
    "E00700",  # State income tax refunds
    "E00800",  # Alimony received
    "E00900",  # Business or profession (Schedule C) net profit/loss (+/-)
    "E01000",  # Net capital gain or loss (+/-)
    "E01100",  # Capital gain distributions reported on Form 1040
    "E01200",  # Other gains (or loss) (+/-)
    "E01400",  # Taxable IRA distribution
    "E01500",  # Total pensions and annuities received
    "E01700",  # Pensions and annuities included in AGI
    "E02000",  # Schedule E net income or loss (+/-)
    "E02100",  # Schedule F net profit/loss (+/-)
    "E02300",  # Unemployment compensation in AGI
    "E02400",  # Gross Social Security benefits
    "E02500",  # Social Security benefits in AGI
]

puf_data = puf[INPUT_VARIABLES + OUTPUT_VARIABLES]

puf_data = puf_data[puf_data.MARS != 0]

st.subheader("Training data")

with st.expander("PUF training data"):
    st.dataframe(puf_data)

with st.expander("CPS data in PUF format for imputation"):
    st.dataframe(cps_data)

from survey_enhance import Imputation


@st.cache_resource
def load_model():
    IMPUTE = False

    if IMPUTE:
        income = Imputation()
        income.train(
            puf_data[INPUT_VARIABLES],
            puf_data[OUTPUT_VARIABLES],
            num_trees=100,
        )
        income.save("income.pkl")
    else:
        income = Imputation.load("income.pkl")

    return income


income = load_model()

# cps_imputation_predictions = income.predict(cps_data[INPUT_VARIABLES])

# imputed_cps = pd.concat([cps_data.copy(), cps_imputation_predictions], axis=1)

st.subheader("Household estimator")

col1, col2 = st.columns(2)

with col1:
    mars = st.selectbox(
        "MARS",
        [
            "SINGLE",
            "JOINT",
            "SURVIVING_SPOUSE",
            "SEPARATE",
            "HEAD_OF_HOUSEHOLD",
        ],
    )
    child_dependents = st.number_input(
        "Child dependents", min_value=0, max_value=10, value=0
    )
    adult_dependents = st.number_input(
        "Adult dependents", min_value=0, max_value=10, value=0
    )


def get_predictions(input_data, num_predictions):
    predictions = []
    for step in range(num_predictions):
        predictions.append(income.predict(input_data))

    prediction_df = pd.concat(predictions)
    return prediction_df


with col2:
    mars = RENAMES["filing_status"][mars]

    progress_bar = st.progress(0)
    prediction_df = get_predictions(
        [[mars, child_dependents, adult_dependents]], 100
    )

    import plotly.express as px

    variable = st.selectbox("Variable", OUTPUT_VARIABLES)

    distribution = px.histogram(
        prediction_df[variable], nbins=300
    ).update_layout(
        title=f"{variable} distribution",
        xaxis_title=variable,
        yaxis_title="Count",
        xaxis_tickformat="$",
        showlegend=False,
    )

    st.plotly_chart(distribution)
